#!/usr/bin/env python
# -*- coding: utf-8 -*-
from array import array
from hashlib import md5
from zorm.config import db_by_table, mc
from zorm.model import _key_by_table

class Kv(object):
    def __init__(self, table, NULL=''):
        self.__table__ = table
        db = db_by_table(table)
        self.cursor = cursor = db.cursor()
        mc_key = _key_by_table(db, table, cursor) 
        self.__mc_id__ = "%s'+%%s"%mc_key
        self.NULL = NULL


    def get(self, id):
        mc_key = self.__mc_id__ % id
        r = mc.get(mc_key)
        if r is None:
            cursor = self.cursor
            cursor.execute('select value from %s where id=%%s' % self.__table__, id)
            r = cursor.fetchone()
            if r:
                r = r[0]
            if r is None:
                r = self.NULL
            mc.set(mc_key, r)
        return r

    def get_dict(self, id_list):
        if type(id_list) not in (array, list, tuple, dict):
            id_list = tuple(id_list)
        mc_key = self.__mc_id__
        result = mc.get_dict([mc_key%i for i in id_list])
        r = {}
        for i in id_list:
            t = result.get(mc_key%i)
            if t is None:
                t = self.get(i)
            r[i] = t
        return r

    def get_list(self, id_list):
        if type(id_list) not in (array, list, tuple, dict):
            id_list = tuple(id_list)
        mc_key = self.__mc_id__
        result = mc.get_dict([mc_key%i for i in id_list])
        r = []
        for i in id_list:
            t = result.get(mc_key%i)
            if t is None:
                t = self.get(i)
            r.append(t)
        return r

    def bind(self, li, property, key='id'):
        for i, value in zip(li, self.get_list(getattr(i,key) for i in li)):
            i.__dict__[property] = value

    def iteritems(self):
        id = 0
        cursor = self.cursor
        while True:
            cursor.execute('select id,value from %s where id>%%s order by id limit 128' % self.__table__, id)
            result = cursor.fetchall()
            if not result:
                break
            for id, value in result:
                yield id, value


    def set(self, id, value):
        r = self.get(id)
        if r != value:
            mc_key = self.__mc_id__ % id
            cursor = self.cursor
            table = self.__table__
            cursor.execute(
                'insert into %s (id,value) values (%%s,%%s) on duplicate key update value=values(value)' % table,
                (id, value)
            )
            cursor.connection.commit()
            if value is None:
                value = False
            mc.set(mc_key, value)

    def mc_flush(self, id):
        mc_key = self.__mc_id__ % id
        mc.delete(mc_key)

    def delete(self, id):
        cursor = self.cursor
        cursor.execute('delete from %s where id=%%s' % self.__table__, id)
        mc_key = self.__mc_id__ % id
        mc.delete(mc_key)

    def id_by_value(self, value):
        cursor = self.cursor
        cursor.execute(
            'select id from %s where value=%%s' % self.__table__,
            value
        )
        r = cursor.fetchone()
        if r:
            r = r[0]
        else:
            r = 0
        return r




#coding:utf-8
from zorm.config import DEBUG
from time import time
import logging
from operator import itemgetter
from MySQLdb import escape_string

itemgetter0 = itemgetter(0)

def _is_null(kwargs, _condition_str):
    kwds = {}
    for k, v in kwargs.iteritems():
        if v is None:
            _condition_str.append('`%s` is NULL'%k)
        else:
            kwds[k] = v
    return kwds

class Query(object):

    def __init__(self, query_type='SELECT *', args=(), conditions={}, model=None, db=None):
        #from zorm.model import Model
        self.type = query_type
        if args:
            self._condition_str = [args[0]]
            self._conditions_para = list(args[1:])
        else:
            self._condition_str = []
            self._conditions_para = []

        self.conditions = _is_null(conditions, self._condition_str)

        self.order = ''
        self.limit = ()
        self.cache = None
        self.model = model

        if db:
            self.__db__ = db
        elif model:
            self.__db__ = model.__db__

    def __getitem__(self, k):
        if self.cache != None:
            return self.cache[k]

        if isinstance(k, (int, long)):
            self.limit = (k, 1)
            lst = self._data()
            if not lst:
                return None
            return lst[0]
        elif isinstance(k, slice):
            if k.start == 0 and  k.stop is None:
                self.limit = ()
            elif k.start is not None:
                assert k.stop is not None, 'Limit must be set when an offset is present'
                assert k.stop >= k.start, 'Limit must be greater than or equal to offset'
                self.limit = k.start, (k.stop - k.start)
            elif k.stop is not None:
                self.limit = 0, k.stop

        return self._data()

    def __len__(self):
        return len(self._data())

    def __iter__(self):
        return iter(self._data())

    def __repr__(self):
        return repr(self._data())

    def count(self, what='*'):
        if self.cache is None:
            result = Query.execute(
                'SELECT COUNT(%s) FROM %s %s' % (
                    what,
                    self.model.__table__,
                    self._keys() or ''
                ),
                self._values(),
                self.__db__
            )

            #奇怪,DBUtils不这样写返回的就是None
            result = result.fetchone()
            if result:
                result = result[0]
            else:
                result = 0
        else:
            result = len(self.cache)

        return result



    def where(self, *args, **kwargs):
        if args:
            self._condition_str.append(args[0])
            self._conditions_para.extend(args[1:])
        kwds = _is_null(kwargs, self._condition_str)
        self.conditions.update(kwds)

        return self

    def order_by(self, field):
        self.order = 'ORDER BY %s' % field
        return self

    def _keys(self):
        if len(self.conditions) or len(self._condition_str):
            return 'WHERE %s' % ' AND '.join(
                ['`%s`=%%s' % k for k in self.conditions]+self._condition_str
            )

    def _values(self):
        return list(self.conditions.itervalues())+self._conditions_para

    def _query(self):
        return '%s FROM %s %s %s %s' % (
            self.type,
            self.model.__table__,
            self._keys() or '',
            self.order,
            self._limit() or '',
        )

    def _limit(self):
        if len(self.limit):
            return 'LIMIT %s' % ','.join(str(l) for l in self.limit)

    def _data(self):
        if self.cache is None:
            self.cache = list(self.iterator())
        return self.cache

    def iterator(self):
        q = self._execute()
        q = q.fetchall()
        for row in q:
            obj = self.model(*row)
            obj._new_record = False
            yield obj

    def _execute(self):
        values = self._values()
        sql = self._query()
        return Query.execute(sql, values, self.__db__)

    def update(self, *args, **kwds):
        values = self._values()
        update_set = []
        if args:
            update_set.append(args[0])
            values = args[1:]+values
        if kwds:
            update_set.append(
                ','.join(
                    '`%s`=%%s'%k for k in kwds.keys()
                )
            )
            values = list(kwds.values())+values
        if update_set:
            Query.execute(
                'UPDATE %s SET %s %s' % (
                    self.model.__table__,
                    ','.join(update_set),
                    self._keys() or ''
                ),
                values,
                self.__db__
            )

    def delete(self):
        values = self._values()
        Query.execute(
            'DELETE FROM %s %s' % (
                self.model.__table__,
                self._keys() or ''
            ),
            values,
            self.__db__
        )


    @classmethod
    def execute(cls, sql, values, db):
        cursor = db.cursor()
        if DEBUG:
            begin_time = time()
        try:
            cursor.execute(sql, values)
        except Exception, ex:
            logging.error(
                '%s\n%s\nException: %s'%(
                    sql, values, ex
                )
            )
            raise
        if DEBUG:
            if type(values) not in (list , tuple):
                values = ( values , )

            logging.info(
                 '%.2f\n\t%s ;'%(
                    1000*(time() - begin_time),
                    sql.strip()%tuple(map(escape_string,values))
                )
            )
        return cursor

    def column(self, column='id', limit=None, offset=None, ):
        self.type = 'SELECT %s' % column
        self.limit = _limit = []
        if limit is not None:
            if offset is not None:
                _limit.append(offset)
            _limit.append(limit)
        r = self._execute().fetchall()

        if r and len(r[0]) == 1:
            return map(itemgetter0,r)

        return r







import os
import msgpack
class SimpleCached:
    ' cache obj in local process, wrapper for memcache '
    def __init__(self, mc):
        self.mc = mc

    def set_msgpack(self, key, val, time=0, dumps=msgpack.packb):
        val = dumps(val)
        self.mc.set(key, val, time)

    def get_msgpack(self, key, loads=msgpack.unpackb):
        r = self.mc.get(key)
        if r is not None:
            r = loads(r)
        return r

    def get_dict_msgpack(self, keys, loads=msgpack.packb):
        rs = self.mc.get_multi(keys)
        rs = dict((k, loads(v)) for k, v in rs.iteritems())
        return rs

    def get_dict(self, keys):
        return self.mc.get_multi(keys)

    def get_list_msgpack(self, keys, loads=msgpack.unpackb):
        rs = self.get_dict_msgpack(keys, loads)
        return [rs.get(k) for k in keys]

    def __getattr__(self, name):
        def func(*args, **kwargs):
            return getattr(self.mc, name)(*args, **kwargs)
        return func

    def reset(self):
        pass

class LocalCached:
    ' cache obj in local process, wrapper for memcache '
    
    def __init__(self, mc):
        self.mc = mc
        self.reset()

    def reset(self):
        self.dataset = {}
        self.dataset_msgpack = {}

    def start_log(self):
        mc = self.mc
        if not hasattr(mc,"obj"):
            self.mc = self.logger(mc)

    def stop_log(self):
        mc = self.mc
        if hasattr(mc,"obj"):
            self.mc = mc.obj

    def get_log(self):
        from collections import defaultdict
        d = defaultdict(int)
        nd = defaultdict(lambda: [0, 0])
        for call, ncall, cost in self.mc.log:
            d[call] += 1
            x = nd[ncall]
            x[0] += 1
            x[1] += cost
        result = 'memcache access (%s/%s calls):\n\n%s\nDetail:\n\n%s\n' % \
                        (len(d), sum(d.itervalues()),
                         ''.join('%s: %d times, %f seconds\n' % (
                                                ncall, times, cost)
                                 for ncall, (times, cost)
                                 in sorted(nd.iteritems())),
                         ''.join('%s: %d times\n' % (key, n)
                                 for key, n in sorted(d.iteritems())))
        result = result.replace('get_multi', 'get_dict')
        return result

    def __repr__(self):
        return 'Locally Cached ' + str(self.mc)

    def set(self, key, val, time=0):
        self.dataset[key] = val
        self.mc.set(key, val, time)

    def set_msgpack(self, key, val, time=0, dumps=msgpack.dumps):
        self.dataset_msgpack[key] = val
        val = dumps(val)
        self.mc.set(key, val, time)

    def get(self, key):
        r = self.dataset.get(key)
        if r is None:
            r = self.mc.get(key)
            if r is not None:
                self.dataset[key] = r
        return r

    def get_msgpack(self, key, loads=msgpack.loads):
        r = self.dataset_msgpack.get(key)
        if r is None:
            r = self.get(key)
            if r is not None:
                r = loads(r)
                self.dataset_msgpack[key] = r
        return r

    def get_dict(self, keys):
        if keys:
            rs = [(k, self.dataset.get(k)) for k in keys]
            r = dict((k, v) for k, v in rs if v is not None)
            rs = self.mc.get_multi([k for k, v in rs if v is None])
            r.update(rs)
            self.dataset.update(rs)
            return r
        else:
            return {}

    def get_dict_msgpack(self, keys, loads=msgpack.loads):
        rs = [(k, self.dataset_msgpack.get(k)) for k in keys]
        r = dict((k, v) for k, v in rs if v is not None)
        rs = self.mc.get_multi([k for k, v in rs if v is None])
        rs = dict((k, loads(v)) for k, v in rs.iteritems())
        r.update(rs)
        self.dataset_msgpack.update(rs)
        return r

    def get_list_msgpack(self, keys, loads=msgpack.loads):
        rs = self.get_dict_msgpack(keys, loads)
        return [rs.get(k) for k in keys]

    def get_list(self, keys):
        if keys:
            rs = self.get_dict(keys)
            return [rs.get(k) for k in keys]
        else:
            return []
    def append(self, key, val):
        self.dataset.pop(key, None)
        return self.mc.append(key, val)

    def prepend(self, key, val):
        self.dataset.pop(key, None)
        return self.mc.prepend(key, val)

    def delete(self, key):
        self.dataset.pop(key, None)
        return self.mc.delete(key)

    def decr(self, key, val=1):
        self.dataset.pop(key, None)
        return self.mc.decr(key, val)

    def incr(self, key, val=1):
        self.dataset.pop(key, None)
        return self.mc.incr(key, val)

    def __getattr__(self, name):
        def func(*args, **kwargs):
            return getattr(self.mc, name)(*args, **kwargs)
        return func



def init_mc(memcached_addr=None, disable_local_cached=False):

    import cmemcached
    kw = {}
    kw['comp_threshold'] = 4096

    #kw['dist'] = cmemcached.DIST_CONSISTENT_KETAMA

    mc = cmemcached.Client(memcached_addr, **kw)

    if disable_local_cached:
        mc = SimpleCached(mc)
    else:
        mc = LocalCached(mc)

    return mc





#coding:utf-8
from intstr import IntStr

redis_keyer = IntStr(
'!"#$&()+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ^_`abcdefghijklmnopqrstuvwxyz{|}~'
)

REDIS_KEY_ID = 'RedisKeyId'
REDIS_KEY = 'RedisKey'
REDIS_ID_KEY = 'RedisIdKey'

_EXIST = set()

class RedisKey(object):
    def __init__(self, redis):
        self.redis = redis

    def __getattr__(self, attr):
        redis = self.redis
        def _(name=''):
            key = attr+name
            if key in _EXIST:
                raise Exception('redis key is already defined %s'%key)
            _EXIST.add(key)
            if redis:
                _key = redis.hget(REDIS_KEY, key)
                if _key is None:
                    id = redis.incr(REDIS_KEY_ID)
                    _key = redis_keyer.encode(id)
                    if name and "%" in name:
                        _key = _key+"'"+name
                    p = redis.pipeline()
                    p.hset(REDIS_KEY, key, _key)
                    p.hset(REDIS_ID_KEY, _key, key)
                    p.execute()
                return _key

        return _


from zorm.config import redis
redis_key = RedisKey(redis)

if __name__ == '__main__':
    pass



# -*- coding: UTF-8 -*-
import time
from warnings import warn
import MySQLdb
from _mysql_exceptions import Warning, InterfaceError, DataError,\
     DatabaseError, OperationalError, IntegrityError, InternalError,\
     NotSupportedError, ProgrammingError
from MySQLdb.converters import FIELD_TYPE, conversions
from config import THREAD_SAFE

if THREAD_SAFE:
    from DBUtils.PersistentDB  import PersistentDB as DB
    def connection(*args, **kwds):
        kwds['maxusage'] = False
        persist = DB (MySQLdb, *args, **kwds)
        conn = persist.connection()
        return conn
else:
    from DBUtils.SteadyDB import connect
    def connection(*args, **kwds):
        kwds['maxusage'] = False
        return connect(MySQLdb, *args, **kwds)


class LogCursor(object):
    def __init__(self, cursor):
        self._cursor = cursor
        self.log = []

    def execute(self, *a, **kw):
        t1 = time.time()
        try:
            retval = self._cursor.execute(*a, **kw)
        except:
            self.log.append((a, kw, 0))
            raise
        timecost = time.time() - t1
        self.log.append((a, kw, timecost))
        return retval

    def __iter__(self):
        return iter(self._cursor)

    def __getattr__(self, attr):
        return getattr(self._cursor, attr)

class SqlFarm:
    def __init__(self, conf=None, **kwargs):
        self.dbcnf = self.parse_config_string(conf)
        self.dbcnf.update(kwargs)
        self._cursor = None
        self.b_commit = True

    def connect(self, host, user, passwd, db, **kwargs):
        conn_params = dict(host=host, user=user,
                db=db, init_command='set names utf8',
                **kwargs)
        if passwd :
            conn_params['passwd'] = passwd
        conv = conversions.copy()
        conv.update({
             FIELD_TYPE.TIMESTAMP: None,
             FIELD_TYPE.DATETIME: None,
             FIELD_TYPE.TIME: None,
             FIELD_TYPE.DATE: None,
        })

        conn_params['conv'] = conv
        conn = connection(**conn_params)

        if not conn:
            raise DatabaseError('can not connect to database: %s %s %s'
                         % (host, user, db))
        cursor = conn.cursor()
        #cursor.execute('set sort_buffer_size=2000000')
        cursor = CursorWrapper(cursor, self)
        return cursor

    def close(self):
        if self._cursor:
            self._cursor.connection.close()

    def cursor(self, ro=False):
        if self._cursor is None:
            self._cursor = self.connect(**self.dbcnf)
        return self._cursor

    def start_log(self):
        if self._cursor is None:
            self._cursor = self.connect(**self.dbcnf)
        self._cursor = LogCursor(self._cursor)

    def stop_log(self):
        if self._cursor is not None:
            self._cursor = self._cursor._cursor

    def get_log(self, name):
        def sql_log(name, log):
            if log:
                return '%s: %d SQL statements (%s seconds):\n%s\n\n' % (
                        name, len(log), sum(x[2] for x in log),
                        '\n'.join(['%8.6fsec %s' % (timecost, a)
                                  for a, kw, timecost in log]))
            else:
                return '%s No Sql Log\n\n'%name
        so = sql_log(name, self._cursor.log)
        return so

    def parse_config_string(self, s):
        dummy = s.split(':')
        if len(dummy) == 4:
            host, db, user, passwd = dummy
            return dict(host=host, db=db, user=user, passwd=passwd)
        elif len(dummy) == 5:
            host, port, db, user, passwd = dummy
            return dict(host=host, port=int(port), db=db, user=user,
                    passwd=passwd)
        else:
            raise ValueError(s)

class SqlStore:

    def __init__(self, host='', port='', user='', password='', db='',
                 db_config=None, **kwargs):
        self.farms = {}
        self.tables = {}
        if db_config is not None:
            for name, f in db_config.items():
                farm = SqlFarm(f['master'], **kwargs)
                self.farms[name] = farm
                for table in f['tables']:
                    self.tables[table] = farm
            #if db_config and '*' not in self.tables:
                #raise DatabaseError('No default farm specified')

        else:
            farm = SqlFarm(':'.join([host, port, db, user, password]))
            self.farms[db] = farm
            self.tables['*'] = farm

        # cache. SqlStore needs to be reloaded is any of these is updated
        self._categories = {}
        self._category_list = None

    def close(self):
        for farm in self.farms.values():
            farm.close()

    def get_farm(self, farm_name):
        farm = self.farms.get(farm_name)
        if farm is None:
            warn('Farm %r is not configured, use default farm' % farm_name,
                    stacklevel=3)
            return self.tables['*']
        else:
            return farm

    def get_db_by_table(self, table):
        farm = self.tables.get(table)
        if farm is None:
            raise Exception("%s TABLE NOT CONFIGURE" % table)
        else:
            return farm

    def cursor(self, ro=False, farm=None, table='*', tables=[]):
        """get a cursor according to table or tables.

        Note:

          * If `tables` is given, `table` is ignored.
          * If `farm` is given, `table` and `tables` are both ignored.
        """

        if farm:
            farm = self.get_farm(farm)
        elif tables:
            farms = set(self.get_db_by_table(table) for table in tables)
            if len(farms) > 1:
                raise DatabaseError('%s are not in the same farm' % (tables, ))
            farm = farms.pop()
        else:
            farm = self.get_db_by_table(table)
        return farm.cursor(ro=ro)

    def start_log(self):
        for farm in self.farms.values():
            farm.start_log()

    def stop_log(self):
        for farm in self.farms.values():
            farm.stop_log()

    def get_log(self):
        r = ' '.join(farm.get_log(name) for name, farm in self.farms.items())
        return r

    def sync_row(self, from_farm, from_table, to_farm, to_table, **conditions):
        """sync rows of a table from one farm to another.

        Eg.
            store.sync_row('b_farm', 'a_farm.entry_vote',
                           'a_farm', 'entry_vote',
                           user_id=1000001, entry_id=10013364)
        The above method call will copy the first row of entry_vote table
        on b_farm to the corresponding table on a_farm.
        """
        try:
            self._sync_row(from_farm, from_table, to_farm, to_table,
                    **conditions)
        except IntegrityError:
            import traceback; traceback.print_exc()

    def _sync_row(self, from_farm, from_table, to_farm, to_table, **conditions):
        cols, vals = zip(*conditions.items())
        from_cursor = self.cursor(farm=from_farm)
        to_cursor = self.cursor(farm=to_farm)
        where_clause = 'where ' + ' and '.join([col+'=%s' for col in cols])

        # acquire lock
        from_cursor.connection.commit()
        sql = 'update %s set %s %s' % (from_table,
                ','.join(['%s=%s'%(col, col) for col in cols]),
                where_clause)
        from_cursor.execute(sql, vals)

        from_cursor.execute('select * from '+from_table+' ' + where_clause, vals)
        rows = from_cursor.fetchall()

        to_cursor.execute('delete from '+to_table+' ' + where_clause, vals)
        if rows:
            sql = 'insert into %s values (%s)' % (to_table,
                    ','.join(['%s'] * len(rows[0])))
            to_cursor.executemany(sql, rows)
        to_cursor.connection.commit()

        # release lock
        from_cursor.connection.rollback()


    def rollback_all(self):
        for farm in self.farms.values():
            cursor = farm.cursor()
            cursor.connection.rollback()

class CursorWrapper() :

    def __init__(self, cursor, farm) :
        self._cursor = cursor
        self.farm = farm

    def __getattr__(self, name) :
        return getattr(self._cursor, name)

    def execute(self, *args, **kwargs) :
        try :
            return self._cursor.execute(*args, **kwargs)
        except MySQLdb.OperationalError, e:
            error_no = e.args[0]
            if 2000 <= error_no < 3000  :
                self.farm._cursor = None
            self._cursor.connection.rollback()
            raise
        except MySQLdb.ProgrammingError, e:
            if e.args[0] == 2014:
                self.farm._cursor = None
            self._cursor.connection.rollback()
            raise
        except MySQLdb.IntegrityError, e:
            self._cursor.connection.rollback()
            raise




#coding:utf-8

from zorm.query import Query, itemgetter0
from zorm.config import db_by_table, mc
from msgpack import packb, unpackb
from hashlib import md5
from intstr import IntStr
from zorm.redis import redis, redis_key

REDIS_ZORM_MC_ID = redis_key.zorm_mc_id()
REDIS_ZORM_ID_MC = redis_key.zorm_id_mc()
REDIS_ZORM_MC = redis_key.zorm_mc()

mc_keyer = IntStr(
    '!"#$&()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[]^_`abcdefghijklmnopqrstuvwxyz{|}~'
)

def mc_key_table(key):
    table = key.split("'", 1)[0]
    return redis.hget(REDIS_ZORM_ID_MC, table)

@classmethod
def _commit(cls):
    db = cls.__db__
    cursor = db.cursor()
    cursor.connection.commit()

def _key_by_table(db, table, cursor):
    db_table = '%s.%s'%(db , table)
    key = [ db_table ]

    if cursor.description:
        for i in cursor.description:
            key.append('%s:%s'%(i[0], i[1]))

    key = ' '.join(key)

    keyd = md5(key).digest()
    mckey = redis.hget(REDIS_ZORM_MC, keyd)

    if not mckey:
        id = redis.incr(REDIS_ZORM_MC_ID)
        mckey = mc_keyer.encode(id)
        p = redis.pipeline()
        p.hset(REDIS_ZORM_MC, keyd, mckey)
        p.hset(REDIS_ZORM_ID_MC, mckey, db_table)
        p.execute()
    #else:
    #    print mc_keyer.decode(mckey)
    return mckey

class _Model(type):
    def __new__(cls, name, bases, attrs):
        base0 = bases[0]
        if base0 is object:
            return super(_Model, cls).__new__(cls, name, bases, attrs)

        new_class = type.__new__(cls, name, bases, attrs)

        new_class.__table__ = table = name
        new_class.__db__ = db = db_by_table(table)


        q = Query.execute('SELECT * FROM %s LIMIT 1' % name, (), db)
        new_class.__column__ = column = map(itemgetter0, q.description)

        if base0 is ModelMc:
            new_class.__key__ = "%s'%%s"%(_key_by_table(
                db.dbcnf['db'], name, q
            ).replace('%', '%%'))
        return new_class

def _delete(self):
    Query.execute(
        'DELETE FROM %s WHERE %%s=%%s' % self.__table__ ,
        self.id,
        self.__db__
    )
    return True



@classmethod
def _where(cls, *args, **kwargs):
    return Query(
        model=cls,
        args=args,
        conditions=kwargs
    )


@classmethod
def _count(cls, *args, **kwargs):
    return Query(
        model=cls,
        args=args,
        conditions=kwargs
    ).count(1)

@classmethod
def _iter(cls, id=0, where=None, limit=500):
    while True:
        if where:
            r = cls.where(where)
        else:
            r = cls

        r = r.where('id>%s', id).order_by('id')
        total = tuple(r[:limit])
        if total:
            for i in total:
                #print i.id
                yield i
            id = total[-1].id
        else:
            break

def __ne__(self, other):
    return not (self == other)

def __eq__(self, other):
    if other is not None:
        sid = self.id
        oid = other.id
        if sid is not None and oid is not None:
            return sid == oid
    return False

@classmethod
def _execute(cls, sql, values=()):
    return Query.execute(sql, values, cls.__db__)

@classmethod
def _max_id(cls):
    c = cls.execute(
        'select max(id) from %s'%cls.__table__
    )
    id = c.fetchone()
    if id:
        return id[0]
    return 0

def __init__(self, *args, **kwargs):
    'Allows setting of fields using kwargs'
    self.__dict__['id'] = None

    self._is_new = True

    for i, arg in enumerate(args):
        self.__dict__[self.__column__[i]] = arg

    for i in self.__column__[len(args):]:
        self.__dict__[i] = kwargs.get(i)

    self.__dict__['_updated'] = set()

def __setattr__(self, name, value):
    dc = self.__dict__
    if name[0] != '_' and name in self.__column__:
        dc_value = dc[name]
        if dc_value is None:
            self._updated.add(name)
        else:
            if value is not None:
                value = type(dc_value)(value)
            if dc_value != value:
                self._updated.add(name)
        dc[name] = value
    else:
        attr = getattr(self.__class__, name, None)
        if attr and hasattr(attr, "fset"):
            attr.fset(self, value)
        else:
            dc[name] = value

def _update(self):
    query = [
        'UPDATE %s SET ' % self.__table__,
        ','.join(['`%s`=%%s'%f for f in self._updated]),
        ' WHERE id=%s '
    ]

    values = [getattr(self, f) for f in self._updated]
    values.append(self.id)

    cursor = Query.execute(" ".join(query), values, self.__db__)

def _insert(self):
    'Uses SQL INSERT to create new record'

    id = self.id
    fields = [
        '`%s`'%f for f in self.__column__
        if id is not None or f != 'id'
    ]
    values = [getattr(self, f, None) for f in self.__column__ if id is not None or f != 'id']


    query = 'INSERT INTO %s (%s) VALUES (%s)' % (
           self.__table__,
           ','.join(fields),
           ','.join(['%s'] * len(fields) )
    )
    cursor = Query.execute(query, values, self.__db__)

    if id is None:
        self.id = cursor.lastrowid

    return True

def _save(self):
    if self._is_new:
        self._insert()
        self._is_new = False
    elif self._updated:
        self._update()
    self._updated.clear()
    return self


@classmethod
def _get_or_create(cls, **kwargs):
    ins = cls.get(**kwargs)
    if ins is None:
        ins = cls(**kwargs)
    return ins


@classmethod
def _get(cls, id=None, **kwargs):
    if id is None:
        if not kwargs:
            return
    else:
        kwargs = {
            'id': id
        }
    q = Query(model=cls, conditions=kwargs)
    q.limit = (0, 1)
    q = q._execute()
    q = q.fetchone()
    if q:
        obj = cls(*q)
        obj.__dict__['_is_new'] = False
        return obj


class Model(object):
    """
能把 mysql 自动映射为 python 的 类 , 比如

mysql数据库中有 User 表 , 结构如下 ::

    CREATE TABLE  `UserMail` (
      `id` bigint(20) unsigned NOT NULL AUTO_INCREMENT,
      `mail` varchar(128) ,
      PRIMARY KEY (`id`)
    ) ENGINE=MyISAM DEFAULT CHARSET=utf8;

Model 定义如下

::

    class UserMail(Model):
        pass


#. 新插入一行 ::

    mail = UserMail(mail="zsp007@gmail.com")
    mail.save()
    
    等价于 SQL ::
        
        insert into UserMail (mail) values ("zsp007@gmail.com")

#. 根据id , 获取一行 ::
   
    mail = UserMail.get(12)
    
    等价于 SQL ::

        select * from UserMail where id = 12 
 
#. 插入或修改一行 ::
    
        user_mail = UserMail.get_or_create(id=21, mail='zsp007@gmail.com')
        user_mail.mail = 'test@example.com'
        user_mail.save()

#. 执行 execute
    
    使用execute方法可以直接执行SQL语句，注意表名大小写 ::

        UserMail.execute('select count(*) from UserMail').fetchone()


#. 删除id为21的行 ::

        UserMail.where(id=21).delete()

    等价于 SQL ::

        DELETE FROM UserMail WHERE id=21

#. 获取最大的id值 ::

        UserMail.max_id()

    等价于 SQL ::

        select max(id) from UserMail

#. 修改一行 ::

    UserMail.where(id=21).update(mail='test@example.com')

    等价于 SQL ::

        UPDATE UserMail SET `mail`=%s WHERE `id`=%s     ['test@example.com', 21]

    """


    __metaclass__ = _Model

    __ne__ = __ne__
    __eq__ = __eq__
    __init__ = __init__

    execute = _execute
    max_id = _max_id
    count = _count
    iter = _iter
    where = _where
    __setattr__ = __setattr__
    save = _save
    get = _get
    _insert = _insert
    get_or_create = _get_or_create

    delete = _delete
    _update = _update

    commit = _commit



def _dumps(self):
    dict = self.__dict__

    value = tuple(
        dict.get(i, None)
        for i in self.__column__
    )

    return packb(value)

class ModelMc(object):

    __metaclass__ = _Model

    __ne__ = __ne__
    __eq__ = __eq__
    __init__ = __init__
    execute = _execute
    max_id = _max_id
    count = _count
    iter = _iter
    where = _where
    __setattr__ = __setattr__
    _insert = _insert
    save = _save
    get = _get
    get_or_create = _get_or_create

    @classmethod
    def _loads(cls, value):
        value = unpackb(value)
        value = cls(*value)
        value._is_new = False
        return value

    @classmethod
    def mc_bind(cls, li, property, key='id'):
        d = []
        e = []
        for i in li:
            k = getattr(i, key)
            if k:
                d.append(k)
                e.append((k, i))
            else:
                i.__dict__[property] = None

        r = cls.mc_get_dict(set(d))
        for k, v in e:
            v.__dict__[property] = r.get(k)




    def delete(self):
        _delete(self)
        self.mc_flush()

    def mc_flush(self):
        mc.delete(self.__key__%self.id)

    @classmethod
    def mc_delete(cls, id):
        mc.delete(cls.__key__%id)

    def _update(self):
        _update(self)
        self.mc_set()


    def mc_set(self):
        key = self.__key__%self.id
        mc.set_msgpack(key, self, dumps=_dumps)

    @classmethod
    def mc_get(cls, id):
        if id:
            key = cls.__key__%id
            value = mc.get_msgpack(key, cls._loads)
            if value is None:
                value = cls.get(id)
                if value:
                    value.mc_set()
            return value

    @classmethod
    def mc_get_dict(cls, id_list):
        if type(id_list) not in (array, list, tuple, dict):
            id_list = tuple(id_list)
        mc_key = cls.__key__
        result = mc.get_dict_msgpack([mc_key%i for i in id_list], cls._loads)
        r = {}
        for i in id_list:
            t = result.get(mc_key%i)
            if t is None:
                if i:
                    t = cls.get(i)
                    if t:
                        t.mc_set()
            r[i] = t
        return r

    @classmethod
    def mc_get_list(cls, id_list):
        if type(id_list) not in (array, list, tuple, dict):
            id_list = tuple(id_list)
        mc_key = cls.__key__
        result = mc.get_dict_msgpack([mc_key%i for i in id_list], cls._loads)
        r = []
        for i in id_list:
            t = result.get(mc_key%i)
            if t is None:
                if i:
                    t = cls.get(i)
                    if t:
                        t.mc_set()
            r.append(t)
        return r
    
    commit = _commit


if __name__ == '__main__':
    pass




#coding:utf-8
from setuptools import setup, find_packages


setup(
    name='zorm',
    version="0.111",
    description="A minimal ORM",
    author="zuroc 张沈鹏",
    author_email="zsp042@gmail.com",
    packages = ['zorm'],
    install_requires = [
        'msgpack-python',
        'MySQL-python',
        'hiredis',
        'redis',
        'DBUtils',
        'intstr',
        'pyrex',
        'decorator',
    ],

)




#coding:utf-8
import os
from decorator import decorator
import inspect
from array import array
from struct import pack
from config import mc

def _mc_get_dict(self, args_list):
    if not isinstance(args_list, (list, tuple, dict, set)):
        args_list = tuple(args_list)
    if args_list:
        result = self.get_list(args_list)
        return dict(zip(args_list, result))
    return {}

def _mc_decorator(self, key=None, expire=0):
    if type(key) is str:
        _key = key
        key = lambda x:_key.format(**x)
    elif key is None:
        key = lambda x:''

    def _func(func):
        arg_names, varargs, varkw, defaults = inspect.getargspec(func)

        if varargs or varkw:
            raise Exception('do not support varargs')

        defaults = defaults or {}
        if defaults:
            args = dict(zip(arg_names[-len(defaults):], defaults))
        else:
            args = {}


        def _(f, *a, **kw):
            aa = args.copy()
            aa.update(zip(arg_names, a))
            aa.update(kw)
            mc_key = key(aa)

            #print mc_key
            r = self.get(mc_key)
            if r is None:
                r = f(*a, **kw)
                self.set(mc_key, r, expire)
            return r

        return decorator(_, func)

    return _func

def _mc_delete(self, *args):
    key = self.key_pattern%args
    return mc.delete(key)

class McCacheM(object):
    def __init__(self, key_pattern):
        self.key_pattern = key_pattern

    def get(self, *args):
        key = self.key_pattern%args
        return mc.get_msgpack(key)

    def set(self, key, value, expire=0):
        return mc.set_msgpack(self.key_pattern%key, value, expire)

    def get_list(self, args_list):
        key_pattern = self.key_pattern
        key_list = [key_pattern%i for i in args_list]
        result = mc.get_list_msgpack(key_list)
        return result

    get_dict = _mc_get_dict
    __call__ = _mc_decorator
    delete = _mc_delete

class McCache(object):
    """

演示代码如下 ::

    mc_xxx = McCache("XXxxx:%s")

    @mc_xxx(lambda x:x['id'])
    def xxx(id):
        return id*3

    print  xxx("123")

    @mc_xxx("{id}")
    def xxx(id):
        return id*3

    print  xxx("467")
    print "MC GET"
    print mc_xxx.get("123")
    print mc_xxx.get_dict(["123","467"])
    mc_xxx.delete("123")


    """
    def __init__(self, key_pattern):
        self.key_pattern = key_pattern

    def get(self, *args):
        key = self.key_pattern%args
        return mc.get(key)

    def set(self, key, value, expire=0):
        return mc.set(self.key_pattern%key, value, expire)

    def get_list(self, args_list):
        key_pattern = self.key_pattern
        key_list = [key_pattern%i for i in args_list]
        result = mc.get_list(key_list)
        return result

    def decr(self, *args):
        key = self.key_pattern%args
        return mc.decr(key)

    def incr(self, *args):
        key = self.key_pattern%args
        return mc.incr(key)

    __call__ = _mc_decorator
    get_dict = _mc_get_dict
    delete = _mc_delete


class McCacheA(object):
    def __init__(self, key_pattern, type='L'):
        self.key_pattern = key_pattern
        self.type = type

    def get(self, *args):
        key = self.key_pattern%args
        result = mc.get(key)
        if result is not None:
            return array(self.type, result)

    def set(self, key, value, expire=0):
        key = self.key_pattern%key
        if type(value) is not array:
            value = pack(self.type*len(value), *value)
        else:
            value = value.tostring()
        return mc.set(key, value, expire)

    def get_list(self, args_list):
        if args_list:
            key_pattern = self.key_pattern
            key_list = [key_pattern%i for i in args_list]
            result = []
            for i in mc.get_list(key_list):
                if i is None:
                    result.append(i)
                else:
                    result.append(array(self.type).fromstring(i))
            return result
        else:
            return []

    __call__ = _mc_decorator
    get_dict = _mc_get_dict
    delete = _mc_delete


class McLimitA(object):
    McCls = McCacheA
    def __init__(self, key_pattern, limit):
        self.mc = self.McCls(key_pattern)
        self.limit = limit
        self.key_pattern = key_pattern

    def __call__(self, key, expire=0):
        if type(key) is str:
            _key = key
            key = lambda x:_key.format(**x)

        def _func(func):
            arg_names, varargs, varkw, defaults = inspect.getargspec(func)

            if varargs or varkw:
                raise Exception('do not support varargs')

            defaults = defaults or {}
            if defaults:
                args = dict(zip(arg_names[-len(defaults):], defaults))
            else:
                args = {}


            def _(f, *a, **kw):
                aa = args.copy()
                aa.update(zip(arg_names, a))
                aa.update(kw)
                mc_key = key(aa)
                offset = aa.get('offset', 0)
                limit = aa.get('limit')

                if limit is not None and offset+limit > self.limit:
                    return func(*a, **kw)

                smc = self.mc
                r = smc.get(mc_key)
                #print  mc_key,r
                if r is None:
                    aa['offset'] = 0
                    aa['limit'] = self.limit
                    r = f(**aa)
                    smc.set(mc_key, r, expire)

                if limit is None:
                    if self.limit > len(r):
                        return r[offset:]
                    return func(*a, **kw)

                return r[offset:limit+offset]

            return decorator(_, func)
        return _func


    def delete(self, *args):
        return self.mc.delete(*args)


class McLimit(McLimitA):
    McCls = McCache


class McLimitM(McLimitA):
    McCls = McCacheM


class McNum(object):
    def __init__(self, get_num, mc_key, timeout=36000):
        self.mc_key = mc_key
        self.get_num = get_num
        self.timeout = timeout


    def __call__(self, *key):
        mk = self.mc_key % '_'.join(map(str, key))
        num = mc.get(mk)
        if num is None:
            #print "self.get_num", key, self.get_num(*key)
            num = self.get_num(*key) or 0
            mc.set(mk, num, self.timeout)
        return num

    def get_dict(self, keys):
        mc_key = self.mc_key
        mc_key_list = dict([(key, mc_key%key) for key in keys])
        result = mc.get_dict(mc_key_list.itervalues())
        r = {}
        for k, mck in mc_key_list.iteritems():
            v = result.get(mck)
            if v is None:
                v = self.get_num(k) or 0
                mc.set(mck, v)
            r[k] = v
        return r

    def bind(self, xxx_list, property, key='id'):
        d = []
        e = []
        for i in xxx_list:
            k = getattr(i, key)
            if k:
                d.append(k)
                e.append((k, i))
            else:
                i.__dict__[property] = None

        r = self.get_dict(set(d))
        for k, v in e:
            v.__dict__[property] = r.get(k)

    def delete(self, *key):
        mk = self.mc_key % '_'.join(map(str, key))
        mc.delete(mk)
    
    def set(self, key, value):
        mk = self.mc_key % key
        mc.set(mk, value)

    def get_list(self, keys):
        r = self.get_dict(keys)

        return [
            r.get(i, 0) for i in keys
        ]

    def incr(self, *key):
        mk = self.mc_key % '_'.join(map(str, key))
        if mc.get(mk) is not None:
            mc.incr(mk)

    def decr(self, *key):
        mk = self.mc_key % '_'.join(map(str, key))
        if mc.get(mk) is not None:
            mc.decr(mk)
